#import "@preview/lovelace:0.2.0": *
#import "../../lib/admonitions.typ": todo, disclosure
#pagebreak(weak: true)
== ROS Packages

This section details ROS packages that I've built or contributed to that were necessary to build this project.

=== lasr_vision_yolov8

This package provides us with the ability to do object detection and segmentation from any ROS node. It is equivalent to
the object detection node described in the specification, however, to simplify the final implementation of 3D detections
of objects and people, another service was also created that accepts pointcloud information to map the detections into
map space.

As such, the node provides a service `/detect` which is implementing according to @data-yolo-detection, and it provides
another service `/detect3d` which is implemented according to the data structure below.

#figure(
  kind: "data",
  supplement: [Data Structure],
  caption: "YOLO 3D Detection Service",
  grid(
    columns: (auto, auto),
    inset: 5pt,
    table(
      columns: (auto),
      inset: 5pt,
      align: horizon + left,
      table.header([*YoloDetection3D.srv*]),
      [\- pcl: *sensor_msgs/PointCloud2* #linebreak() \- dataset: *string* #linebreak() \- confidence: *float* #linebreak() \- nms:
        *float*],
      [\- detected_objects: *YoloDetection3D[]*],
    ),
    table(
      columns: (auto),
      inset: 5pt,
      align: horizon + left,
      table.header([*YoloDetection3D.msg*]),
      [\- name: *string* #linebreak() \- confidence: *float* #linebreak() \- xywh: *int[]* #linebreak() \- xyseg: *int[]* #linebreak() \- target_frame:
        *string* #linebreak() \- point: *geometry_msgs/Point*],
    ),
  ),
) <data-yolo-service>

Under the hood, it uses the ultralytics @Jocher_YOLO_by_Ultralytics_2023 Python library to load and use YOLOv8 models.
When either of the services is invoked, the first thing that occurs is the *dataset* value is read and the corresponding
model is instantiated. We use a basic caching mechanism here to avoid having to initialise models more than once:

#algorithm(caption: [Basic Cache Algorithm], pseudocode-list[
  - *input:* dataset: _string_
  - *output:* instatiated_model: _Model_
  - *globals:* loaded_models: _int $->$ Model_
  + *if* dataset *in* loaded_models *then*
    + instatiated_model $<-$ loaded_models[dataset] #comment[use previously loaded model]
  + *else*
    + instatiated_model $<-$ *Model*(dataset) #comment[load the model into memory]
    + loaded_models[dataset] $<-$ instatiated_model #comment[save it for later use]
  + *end*
])

The actual detection service works as follows:

#algorithm(
  caption: [YOLO Detection Service],
  pseudocode-list[
    - *input:* request: _YoloDetectionRequest_
    - *output:* response: _YoloDetectionResponse_
    - *globals:* load_model: _string $->$ Model_
    - *requires:* cv2_img #comment[cv2_img is another package described later in this section]
    + image $<-$ cv2_img.message_to_image(request.image) #comment[prepare image]
    + model $<-$ load_model(request.dataset) #comment[load the requested model]
    + result $<-$ model(image, request.confidence, request.nms) #comment[infer †he results]
    + response.detected_objects $<-$ result #comment[return the result from the model]
    - #comment[in the actual implementation, the result out of the model needs to be converted to the response type]
  ],
)

However, in the case of 3D detections, we need to flatten the given pointcloud into a 2D image from the camera's
perspective, we replace the first step in the service as such:

#pseudocode-list[
  + #text(red)[$-$ image $<-$ cv2_img.message_to_image(request.image)] #comment[prepare image]
  + #text(olive)[$+$ image $<-$ cv2_pcl.pcl_to_image(request.pcl)] #comment[convert pointcloud to 2D image]
]

The *cv2_pcl* library provides various utilities for working with pointcloud messages, in this case we only care about
*pcl_to_image* which flattens the pointcloud, and *seg_to_centroid* which converts a pointcloud and given 2D
segmentation mask into a point in 3D space (the centre of the points selected by the segmentation mask).

So to then also determine the actual position of the detected objects we use this additional method provided above.
However, this will put us in the camera coordinate frame, so we must transform it to the map frame using tf (for the
purposes of demonstration, the code below assumes all transforms were gathered at the same time and that the camera
frame is relative to the robot base which is relative to the map frame).

#algorithm(
  caption: [Determining object detection centroid],
  pseudocode-list[
    - *input:* pcl: _PointCloud_, detection: _YoloDetection3D_
    - *output:* point: _Point_
    - *globals:* tf_buffer: _TfBuffer_
    + centroid $<-$ cv2_img.seg_to_centroid(pcl, detection.xyseg) #comment[find centroid]
    + point $<-$ tf_buffer.transform(centroid, "map") #comment[transform point to map frame]
      - #text(gray, [*transform()* will perform the following:]) #comment[(simplified to explain how the transforms work)]
      - *input:* point: _Point_
      - *output:* transformed_point: _Point_
      - *buffer:* $T_c^r, T_r^m$: _matrix transforms from camera to robot, robot to map_
      + $T_c^m <- T_c^r * T_r^m$ #text(gray, [@6556373]) #comment[find the transformation from camera to map]
      + transformed_point $<- T_c^m * $ point #comment[transform the point accordingly]
  ],
)

#figure(kind: image, caption: "YOLO Inference", table(
  columns: (auto, auto),
  inset: 5pt,
  stroke: none,
  table.header("Input", "Output"),
  image("../../images/inference/camera.png", height: 80pt),
  image("../../images/inference/yolo.png", height: 80pt),
))

#pagebreak(weak: true)
=== lasr_vision_torch

This package uses PyTorch to run inference with a custom-trained face segmentation model as a service
`/torch/detect/face_features` defined by @data-torch-service.

#figure(
  kind: image,
  caption: [Inferences on Hugh Laurie (House M.D. @house_md) and Morten Rieger Hannemose (recreation of Lena @hannemoselena) #linebreak() Segmented
    areas in figures: #text(yellow)[face], #text(red)[hair], #text(green)[hat]],
  table(
    columns: (auto, auto, auto, auto),
    inset: 5pt,
    stroke: none,
    table.header("Input", "Output", "Input", "Output"),
    image("../../images/inference/torch_in1.jpg", height: 80pt),
    image("../../images/inference/torch_out1.png", height: 80pt),
    image("../../images/inference/torch_in2.jpg", height: 80pt),
    image("../../images/inference/torch_out2.png", height: 80pt),
  ),
)

#disclosure[Package was developed in collaboration with a colleague on the KCL LASR robotics team. Work as part of this project was
  limited to integrating a pre-trained model into a ROS package that could then be used to provide these inferences.]

=== lasr_vision_deepface

This package provides a service `/deepface/detect` implemented according to @data-deepface-service. The service uses the
DeepFace Python library @deepface-framework to find and represent all faces in a given image as vectors with their
corresponding bounding boxes.

However, unlike the proposed implementation, we first perform alignment which is the process of roughly estimating where
faces exist in the image and generating a smaller bounding box that we can then process with the purpose-built facial
detection models. By default, the library uses the OpenCV Haar Cascade face detection algorithm @haar, which is fast and
efficient but very error-prone. We opt to use mtcnn which is slower but promises "superior accuracy" in theory and
experimentation @7553523.

#figure(kind: image, caption: "mtcnn Inference Result (using DeepFace framework)", table(
  columns: (auto, auto),
  inset: 5pt,
  stroke: none,
  table.header("Input", "Output"),
  image("../../images/inference/camera.png", height: 80pt),
  image("../../images/inference/deepface.png", height: 80pt),
))

After alignment, we use the requested model to generate a vector representing each face. In this project, we opt to use
VGG-Face for all inferences.

#pagebreak(weak: true)
=== lasr_vision_bodypix

This package provides the `/bodypix/detect` service implemented according to @data-bodypix-service, and uses the
tf_bodypix @bodypix-python library to load a pre-trained BodyPix model for use in inference on the given images.

The library outputs masks in the form of a 2D boolean array which is not possible to directly represent as a ROS message
type. We have to flatten the array and then reshape it on the other end. The new service hence is defined as follows:

#figure(
  kind: "data",
  supplement: [Data Structure],
  caption: "Amendments to service",
  table(columns: (auto, auto), inset: 5pt, stroke: none, table(
    columns: (auto),
    inset: 5pt,
    align: horizon + left,
    table.header([*BodyPixDetection.srv*]),
    "[..]",
    [\- masks: *BodyPixMask[]*],
  ), table(columns: (auto), inset: 5pt, align: horizon + left, table.header([*BodyPixMask.msg*]), [
    \- mask: *bool[]* #linebreak() \- shape: *uint32[]*
  ])),
) <data-bodypix-amended-service>

To perform the flattening and reshaping, we use the numpy library:

#algorithm(caption: [Flattening and reshaping data], pseudocode-list[
  - #text(gray, [Given a 2D numpy array *mask*])
  + message.mask $<-$ mask.flatten()
  + message.shape $<-$ mask.shape
  - #text(gray, [Given the mask message *message*])
  + mask $<-$ numpy.array(message.mask).reshape(\*message.shape)
])

#figure(kind: image, caption: "BodyPix Inference", table(
  columns: (auto, auto),
  inset: 5pt,
  stroke: none,
  table.header("Input", "Output"),
  image("../../images/inference/camera.png", height: 80pt),
  image("../../images/inference/bodypix.png", height: 80pt),
))

#pagebreak(weak: true)
=== lasr_vector_database_weaviate

This package provides a way to quickly deploy the Weaviate @weaviate vector database within our containerised
environment and then provide a ROS native way to communicate with it. To deploy Weaviate, the project recommends
installing Docker and using their configuration generator to build and run the database, this includes configuring
modules such as vectorisers from text or images. Unfortunately, this is not feasible for our environment as quite often
we run in unprivileged Linux environments with no option to install Docker or alternatives such as Podman (which would
let us run the actual containers without root!).

Digging around on their source code repositories, you can find each tagged release gets corresponding binary releases
built and uploaded to GitHub @weaviate_releases. However, this would not be enough to run the database. Inspecting the
Dockerfile used to build their images you can find found what sort of arguments are required to start the binary
@weaviate_dockerfile_cmd. The sample compose file also provides what environment variables are needed (one can generate
one in their documentation @weaviate_documentation_docker_compose).

Using all of this information, the provided node in the package performs the following: Download the chosen Weaviate
binary if it does not exist yet (we provide the rosparam `~version` to select the appropriate release). Unpack,
configure and launch the database as a subprocess. We must allocate some unique port on the local device to communicate
between the database and our node so we choose two ports that can also be changed using the rosparams `~http_port` and
`~grpc_port`. We also allow the end user of the package to run multiple databases by changing the name of the database
using the param `~db_name`.

After the subprocess is created, we poll the HTTP API until it is available, after which we create the three services
described earlier in the specification: `/database/vectors/{name}/create_collection`, `../insert`, and `../query`. These
are defined by @data-weaviate-create, @data-weaviate-insert, and @data-weaviate-query respectively. These simply map the
requests from ROS message types to/from the Weaviate API JSON types and send the requests to the Weviate REST API.

=== cv2_img

Quite often when video or images are sent over ROS, they are serialised as single raw frames that are sent to a topic.
However, when dealing with images in Python, one would quite often either have them as an (RGB) Pillow Image or simply
as a (BGR) numpy arr (also referred to as a cv2 image).

This package provides a Python library that simplifies converting between images sent over ROS and the format we want to
use in Python. It provides the following functions:

```python
def pillow_img_to_msg(img: PIL.Image, stamp=None) -> SensorImage:
def cv2_img_to_msg(img: numpy.ndarray, stamp=None) -> SensorImage:
def msg_to_pillow_img(msg: SensorImage) -> PIL.Image:
def msg_to_cv2_img(msg: SensorImage) -> np.ndarray:
```

#pagebreak(weak: true)
=== colour_estimation

When finding descriptions of people's appearance, we take the mean of the colour of certain parts of their body, such as
the torso. To assist with then translating this into natural language, this package provides the following dictionaries
and functions:

```python
# mappings of colour names to RGB values
RGB_COLOURS = { 'red': [255, 0, 0], .. }
RGB_HAIR_COLOURS = { 'midnight black': (9, 8, 6), .. }

# function that takes the requested colour
# and a dictionary defined above
# and outputs a list of the closest colour names
# with their distances to the requested colour
def closest_colours(requested_colour, colours):
```

#algorithm(caption: [Closest Colour Algorithm], pseudocode-list[
  - *input:* $r$ requested_colour: _int[3]_, colours: _str $->$ int[3]_
  - *output:* closest_colours: _(name: string, distance: float)[]_
  + distances $<-$ {}
  + *for* colour_name, $v$ rgb_value $<-$ colours *do*
    + $diff <- v - r$
    + distances[colour_name] $<- |diff|$
  + *end*
  + sorted_colours $<-$ sorted(distances, *key*: item: item.value)
  + closest_colours $<-$ sorted_colours[:3]
])

=== lasr_speech_recognition_whisper

This package provides the action `/transcribe_speech` implemented according to @data-transcription-action. To enable
access to the Whisper speech-to-text model, we can use their Python package @pypi_whisper which handles the process of
downloading, loading, and performing inference on our chosen model. We also use the SpeechRecognition
@pypi_speechrecognition Python package which provides many useful utilities that simplify recording audio, capturing
likely phrases, and adjusting to the environment.

To select the input device, the end user can provide an argument to the transcription node. This can be either the name
or the index of the microphone, found by running the provided `list_microphones.py` script.

#algorithm(caption: [Selecting Microphone], pseudocode-list[
  - *input:* device: _string | int | None_
  - *output:* microphone: _Microphone_
  - *requires:* sr: _SpeechRecognition @pypi_speechrecognition/* @typstyle off */_
    + *if* device *is* None *then*
      + microphone $<-$ sr.Microphone() #comment[try to auto-detect the system microphone]
    + *else if* device.isdigit() *then*
      + microphone $<-$ sr.Microphone(device_index = device) #comment[use the given device index]
    + *else*
      + *for* index, name *in enumerate*(sr.Microphone.list()) *do*
        + *if* device *in* name *then*
          + microphone $<-$ sr.Microphone(device_index = index) #comment[use by name]
        + *end*
      + *end*
    + *end*
])

Afterwards, the service implements the action as follows:

#algorithm(
  caption: [Whisper Service],
  pseudocode-list[
    - *requires:* w: _Whisper @pypi_whisper/* @typstyle off */_, sr: _SpeechRecognition @pypi_speechrecognition/* @typstyle off */_,
      actionlib: _actionlib @ros_actionlib/* @typstyle off */_
    + model $<-$ w.load_model("medium.en")
    + microphone $<-$ select_microphone(args.device) #comment[configure the microphone input]
    + recogniser $<-$ sr.Recogniser() #comment[instantiate the speech recongnition library]
    + recogniser.adjust_for_ambient_noise(microphone) #comment[learn background noise and set levels]
    + actionlib.ActionServer(#text(gray)[execute])
      - #text(gray)[execute()]
      - *input:* goal: _TranscribeSpeechGoal_
      - *output:* result: _TranscribeSpeechResult_
      + phrase $<-$ recogniser.listen(microphone, ...) #comment[start listening for phrase, args control timeouts]
      + wav_data $<-$ phrase.as_wav_data() #comment[WAV is a raw data format that we can feed into models]
      + normalised_data $<-$ float\(wav_data\)$slash 32768.0$ #comment[normalise int16 to float values]
      + result $<-$ model.transcribe(normalised_data) #comment[run inference using Whisper]
  ],
)

#pagebreak(weak: true)
=== lasr_rasa

This package is a relatively straightforward wrapper around the RASA Python package @pypi_rasa, it provides the
`/lasr_rasa/parse` service which implements @data-rasa-service. In order to train our model, we can use the included CLI
@rasa_cli.

To create the receptionist dialog model, we first have to initialise the RASA project using `rosrun lasr_rasa rasa init`
in a new directory where we intend for the project to live. Now, we can define training examples by editing
`data/nlu.yml` relative to the newly created project. We can specify our dialog intents as seen in @nlu-example.

#figure(caption: "NLU training data", align(left, [
```yaml
version: "3.1"

nlu:
  - intent: person_name
    examples: |
      - My name is [Adel](name)
      - [Robin](name)
      - [Poppy](name)
      - ...

  - intent: favourite_drink
    examples: |
      - [Orange juice](drink)
      - [Cola](drink)
      - My favourite drink is [iced tea](drink)
      - ...
```
])) <nlu-example>

Now we can run `rosrun lasr_rasa rasa train nlu` to train on this given dataset. The trained model will be output as a
`.tar.gz` archive in `models` relative to the created project. We can now use the package to start the RASA service
using this new model.

=== lasr_skills

The lasr_skills @lasr-skills package is an effort by the KCL LASR team to produce common SMACH @smach states that can be
re-used across many projects, this includes small building blocks such as navigating to a point, detecting objects,
talking to the guest, and so on.

This project uses and contributes some states from/to this package. These are discussed at greater length in the
following section.